@@ -15,8 +15,18 @@ import seaborn as sns
 from sklearn.preprocessing import OrdinalEncoder
 # from bigholes import HoleFinder
 from tqdm import tqdm
+from tqdm import tqdm
+from numpy.random import rand
+from numpy.random import seed
+import acv_explainers
+import numpy as np
+from sklearn.linear_model import LogisticRegression
+from lightgbm import LGBMClassifier
+from sklearn.ensemble import IsolationForest
+import pickle
 
-
+import warnings
+warnings.filterwarnings('ignore')
 
 def plot_feature_importance(importance, names, model_type, xlabel='SHAP values', title=' '):
     # Create arrays from feature importance and feature names
@@ -1321,3 +1331,447 @@ def mc_approx_rules(rules, rules_data, S_star, strategy='random', maxitr=1000, i
         approx_rec[a, S_star[a], 1] = hallOfFame[-1].U
     return approx_rec
 
+
+def return_xy_cnt(x, y_target, S_bar_set, x_train, y_train, w):
+    """
+    return the observations with y=y_target that fall in the projected leaf of x
+    when we condition given S=S_bar of x.
+    """
+
+    x_train_cnt = []
+    for i, wi in enumerate(w):
+        if wi != 0 and y_train[i] == y_target:
+            x_train_cnt.append(x_train[i].copy())
+
+    x_train_cnt = np.array(x_train_cnt)
+    #     x_train_cnt[:, S_bar_set] = x[S_bar_set]
+    y_train_cnt = np.array(x_train_cnt.shape[0] * [y_target])
+
+    return x_train_cnt, y_train_cnt
+
+
+def return_leaf_cnt(ac_explainer, S_star, x_train_cnt, y_train_cnt, x_train, y_train, pi=0.9):
+    """
+    return the original leaves of the observations with y=y_target that fall in the projected leaf
+    when we condition given S=S_bar of x.
+    """
+    sdp, rules = ac_explainer.compute_sdp_rule(x_train_cnt, y_train_cnt,
+                                               x_train, y_train,
+                                               x_train_cnt.shape[0] * [list(range(x_train.shape[1]))]
+                                               )
+    if np.sum(sdp >= pi) != 0:
+        rules_unique = np.unique(rules[sdp >= pi], axis=0)
+    else:
+        rules_unique = np.expand_dims(rules[np.argmax(sdp)], axis=0)
+
+    r_buf = rules_unique.copy()
+    for i in range(rules_unique.shape[0]):
+        list_ric = [r.copy() for r in r_buf if not np.allclose(r, rules_unique[i])]
+        find_union(rules_unique[i], list_ric, S=S_star)
+
+    return rules_unique
+
+
+def remove_in(ra):
+    """
+    remove A if A subset of B in the list of compatible leaves
+    """
+    for i in range(ra.shape[0]):
+        for j in range(ra.shape[0]):
+            if i != j and np.prod([(ra[i, s, 1] <= ra[j, s, 1]) * (ra[i, s, 0] >= ra[j, s, 0])
+                                   for s in range(ra.shape[1])], axis=0).astype(bool):
+                ra[i] = ra[j]
+    return np.unique(ra, axis=0)
+
+
+def get_compatible_leaf(acvtree, x, y_target, S_star, S_bar_set, w, x_train, y_train, pi=0.9, acc_level=0.9):
+    """
+    Compute the compatible leaves and order given their accuracy
+    """
+    x_train_cnt, y_train_cnt = return_xy_cnt(x, y_target, S_bar_set, x_train, y_train, w)
+
+    compatible_leaves = return_leaf_cnt(acvtree, S_star, x_train_cnt, y_train_cnt, x_train, y_train)
+    compatible_leaves = np.unique(compatible_leaves, axis=0)
+    compatible_leaves = remove_in(compatible_leaves)
+    #     compatible_leaves = np.rint(compatible_leaves)
+    compatible_leaves = np.round(compatible_leaves, 2)
+
+    partition_leaf = compatible_leaves.copy()
+    d = partition_leaf.shape[1]
+    nb_leaf = partition_leaf.shape[0]
+    leaves_acc = []
+    suf_leaf = []
+
+    for i in range(nb_leaf):
+        x_in = np.prod([(x_train[:, s] <= partition_leaf[i, s, 1]) * (x_train[:, s] > partition_leaf[i, s, 0])
+                        for s in range(d)], axis=0).astype(bool)
+        y_in = y_train[x_in]
+        acc = np.mean(y_in == y_target)
+
+        leaves_acc.append(acc)
+
+        if acc >= acc_level:
+            suf_leaf.append(partition_leaf[i])
+
+    best_id = np.argmax(leaves_acc)
+
+    return suf_leaf, partition_leaf, leaves_acc, partition_leaf[best_id], leaves_acc[best_id]
+
+
+def return_counterfactuals(ac_explainer, suf_leaf, S_star, S_bar_set, x, y, x_train, y_train, pi_level):
+    """
+    Compute the SDP of each C_S and return the ones that has sdp >= pi_level
+    """
+    counterfactuals = []
+    counterfactuals_sdp = []
+    counterfactuals_w = []
+
+    for leaf in suf_leaf:
+
+        cond = np.ones(shape=(1, x_train.shape[1], 2))
+        cond[:, :, 0] = -1e+10
+        cond[:, :, 1] = 1e+10
+
+        for s in S_bar_set:
+            cond[:, s, 0] = x[:, s]
+            cond[:, s, 1] = x[:, s]
+
+        cond[:, S_star] = leaf[S_star]
+        sdp, w = ac_explainer.compute_ddp_cond_weights(x, y, x_train, y_train, S=[S_bar_set], cond=cond,
+                                                       pi_level=pi_level)
+        if sdp >= pi_level:
+            counterfactuals.append(cond)
+            counterfactuals_sdp.append(sdp)
+            counterfactuals_w.append(w)
+
+    return np.unique(counterfactuals, axis=0), np.unique(counterfactuals_sdp, axis=0), \
+           np.unique(counterfactuals_w, axis=0)
+
+
+def return_global_counterfactuals(ac_explainer, data, y_data, s_star, n_star, x_train, y_train, w, acc_level, pi_level):
+    """
+    stack all to compute the C_S for each observations
+    """
+    N = data.shape[0]
+    suf_leaves = []
+    counterfactuals_samples = []
+    counterfactuals_samples_sdp = []
+    counterfactuals_samples_w = []
+
+    for i in range(N):
+        suf_leaf, _, _, _, _ = get_compatible_leaf(ac_explainer, data[i], 1 - y_data[i], s_star[i], n_star[i], w[i],
+                                                   x_train, y_train, pi=pi_level, acc_level=acc_level)
+        suf_leaves.append(suf_leaf)
+        #         print(suf_leaf)
+        #         print(np.unique(suf_leaf, axis=0).shape, len(suf_leaf))
+        counterfactuals, counterfactuals_sdp, w_cond = \
+            return_counterfactuals(ac_explainer, suf_leaf, s_star[i], n_star[i], data[i].reshape(1, -1),
+                                   y_data[i].reshape(1, -1), x_train, y_train, pi_level)
+
+        counterfactuals_samples.append(counterfactuals)
+        counterfactuals_samples_sdp.append(counterfactuals_sdp)
+        counterfactuals_samples_w.append(w_cond)
+
+    return counterfactuals_samples, counterfactuals_samples_sdp, counterfactuals_samples_w
+
+
+# Fonction global explanations
+
+def return_ge_counterfactuals(ac_explainer, suf_leaf, S_star, S_bar_set, x, y, x_train, y_train, cond_s, pi_level):
+    counterfactuals = []
+    counterfactuals_sdp = []
+    counterfactuals_w = []
+
+    for leaf in suf_leaf:
+
+        cond = cond_s.copy()
+        cond[:, S_star] = leaf[S_star]
+
+        sdp, w = ac_explainer.compute_ddp_cond_weights(x, y, x_train, y_train, S=[[-6]], cond=cond)
+        if sdp >= pi_level:
+            counterfactuals.append(cond)
+            counterfactuals_sdp.append(sdp)
+            counterfactuals_w.append(w)
+
+    return counterfactuals, counterfactuals_sdp, counterfactuals_w
+
+
+def return_ge_global_counterfactuals(ac_explainer, data, y_data, s_star, n_star, x_train, y_train, w, acc_level, cond,
+                                     pi_level):
+    N = data.shape[0]
+    suf_leaves = []
+    counterfactuals_samples = []
+    counterfactuals_samples_sdp = []
+    counterfactuals_samples_w = []
+
+    for i in range(N):
+        suf_leaf, _, _, _, _ = get_compatible_leaf(ac_explainer, data[i], 1 - y_data[i], s_star[i], n_star[i], w[i],
+                                                   x_train, y_train, pi=pi_level, acc_level=acc_level)
+        suf_leaves.append(suf_leaf)
+
+        counterfactuals, counterfactuals_sdp, w_cond = return_ge_counterfactuals(ac_explainer, suf_leaf, s_star[i], n_star[i],
+                                                                                 data[i].reshape(1, -1),
+                                                                                 y_data[i].reshape(1, -1), x_train,
+                                                                                 y_train,
+                                                                                 np.expand_dims(cond[i], 0),
+                                                                                 pi_level)
+        counterfactuals_samples.append(counterfactuals)
+        counterfactuals_samples_sdp.append(counterfactuals_sdp)
+        counterfactuals_samples_w.append(w_cond)
+
+    return counterfactuals_samples, counterfactuals_samples_sdp, counterfactuals_samples_w
+
+def print_rule(col_name, r, decision=None, sdp=True, output=None):
+    for i, col in enumerate(col_name):
+        if not ((r[i, 0] <= -1e+10 and r[i, 1] >= 1e+10)):
+            print('If {} in [{}, {}] and '.format(col, r[i, 0], r[i, 1]))
+            print(' ')
+
+    if sdp == True:
+        print('Then the output is = {}'.format(output))
+        print('SDP Probability = {}'.format(decision))
+    else:
+        print('Then the output is different from {}'.format(output))
+        print('Counterfactual DDP Probability = {}'.format(decision))
+
+
+def generate_candidate(x, S, x_train, cond, n_iterations):
+    x_poss = [x_train[(cond[i, 0] <= x_train[:, i]) * (x_train[:, i] <= cond[i, 1]), i] for i in S]
+
+    x_cand = np.repeat(x.reshape(1, -1), repeats=n_iterations, axis=0)
+    for i in range(len(S)):
+        rdm_id = np.random.randint(0, x_poss[i].shape[0], n_iterations)
+        x_cand[:, S[i]] = x_poss[i][rdm_id]
+
+    return x_cand
+
+
+def simulated_annealing(outlier_score, x, S, x_train, cond, batch, max_iter, temp,
+                        max_iter_convergence=10):
+    best = generate_candidate(x, S, x_train, cond, 1)
+    best_eval = outlier_score(best)[0]
+    curr, curr_eval = best, best_eval
+
+    move = 0
+    for i in range(max_iter):
+
+        x_cand = generate_candidate(curr, S, x_train, cond, batch)
+        score_candidates = outlier_score(x_cand)
+
+        candidate_eval = np.max(score_candidates)
+        candidate = x_cand[np.argmax(score_candidates)]
+
+        if candidate_eval > best_eval:
+            best, best_eval = candidate, candidate_eval
+            move = 0
+        else:
+            move += 1
+
+        # check convergence
+        if best_eval > 0 and move > max_iter_convergence:
+            break
+
+        diff = candidate_eval - curr_eval
+        t = temp / np.log(float(i + 1))
+        metropolis = np.exp(-diff / t)
+
+        if diff > 0 or rand() < metropolis:
+            curr, curr_eval = candidate, candidate_eval
+
+    return best, best_eval
+
+dataset = 'none'
+
+def save_model(model, name='{}'.format(dataset)):
+    with open('{}.pickle'.format(name), 'wb') as f:
+        pickle.dump(model, f)
+
+
+def load_model(name='{}'.format(dataset)):
+    with open('{}.pickle'.format(name), 'rb') as f:
+        loaded_obj = pickle.load(f)
+    return loaded_obj
+
+
+class RunExperiments:
+
+    def __init__(self, acv_explainer, x_train, x_test, y_train, y_test, columns_name):
+        self.columns_name = columns_name
+        self.acv_explainer = acv_explainer
+        self.x_train = x_train
+        self.x_test = x_test
+        self.y_train = y_train
+        self.y_test = y_test
+        self.ddp_importance_local, self.ddp_index_local, self.size_local, self.ddp_local = None, None, None, None
+        self.S_star_local, self.S_bar_set_local = None, None
+        self.ddp_local, self.w_local = None, None
+        self.counterfactuals_samples_local, self.counterfactuals_samples_sdp_local, \
+        self.counterfactuals_samples_w_local = None, None, None
+        self.isolation = None, None
+        self.dist_local = None
+        self.score_local = None
+        self.errs_local = None
+        self.accuracy_local = None
+        self.coverage_local = None
+        self.sdp_importance_se, self.sdp_index_se, self.size_se, self.sdp_se = None, None, None, None
+        self.S_star_se, self.N_star_se = None, None
+        self.sdp_rules, self.rules, self.sdp_all, self.rules_data, self.w_rules = None, None, None, None, None
+
+        self.ddp_importance_regional, self.ddp_index_regional, self.size_regional, self.ddp_regional = None, None, None, None
+        self.S_star_regional, self.S_bar_set_regional = None, None
+        self.counterfactuals_samples_regional, self.counterfactuals_samples_sdp_regional, \
+        self.counterfactuals_samples_w_regional = None, None, None
+        self.dist_regional = None
+        self.score_regional = None
+        self.errs_regional = None
+        self.accuracy_regional = None
+        self.coverage_regional = None
+
+    def run_local_divergent_set(self, x, y, stop=True, pi_level=0.8):
+        print('### Computing the local divergent set of (x, y)')
+
+        self.ddp_importance_local, self.ddp_index_local, self.size_local, self.ddp_local = \
+            self.acv_explainer.importance_ddp_rf(x, y, self.x_train, self.y_train,
+                                                 stop=stop, pi_level=pi_level)
+
+        self.S_star_local, self.S_bar_set_local = \
+            acv_explainers.utils.get_active_null_coalition_list(self.ddp_index_local, self.size_local)
+        self.ddp_local, self.w_local = self.acv_explainer.compute_ddp_weights(x, y, self.x_train, self.y_train,
+                                                                              S=self.S_bar_set_local)
+
+    def run_local_counterfactual_rules(self, x, y, acc_level=0.8, pi_level=0.8):
+        print('### Computing the local counterfactual rules of (x, y)')
+
+        self.counterfactuals_samples_local, self.counterfactuals_samples_sdp_local, \
+        self.counterfactuals_samples_w_local = return_global_counterfactuals(self.acv_explainer, x, y,
+                                                                             self.S_star_local, self.S_bar_set_local,
+                                                                             self.x_train, self.y_train, self.w_local,
+                                                                             acc_level=acc_level,
+                                                                             pi_level=acc_level)
+
+    def run_sampling_local_counterfactuals(self, x, y, batch=1000, max_iter=1000, temp=0.5):
+        print('### Sampling using the local counterfactual rules of (x, y)')
+
+        self.isolation = IsolationForest()
+        self.isolation.fit(self.x_train)
+        outlier_score = lambda x: self.isolation.decision_function(x)
+
+        self.dist_local = []
+        self.score_local = []
+        self.errs_local = []
+
+        for i in tqdm(range(x.shape[0])):
+            if len(self.counterfactuals_samples_local[i]) != 0:
+                a, sco = simulated_annealing(outlier_score, x[i], self.S_star_local[i], self.x_train,
+                                             self.counterfactuals_samples_local[i][
+                                                 np.argmax(self.counterfactuals_samples_sdp_local[i])][0],
+                                             batch, max_iter, temp)
+                self.dist_local.append(np.squeeze(a))
+                self.score_local.append(sco)
+                self.errs_local.append(
+                    self.acv_explainer.predict(self.dist_local[-1].reshape(1, -1)) != self.acv_explainer.predict(
+                        x[i].reshape(1, -1)))
+        self.accuracy_local = np.mean(self.errs_local)
+        self.coverage_local = len(self.errs_local) / x.shape[0]
+
+    def run_sufficient_rules(self, x_rule, y_rule, pi_level=0.9):
+        print('### Computing the Sufficient Explanations and the Sufficient Rules')
+        self.x_rules, self.y_rules = x_rule, y_rule
+        self.sdp_importance_se, self.sdp_index_se, self.size_se, self.sdp_se = \
+            self.acv_explainer.importance_sdp_rf(x_rule, y_rule,
+                                                 self.x_train, self.y_train,
+                                                 stop=False,
+                                                 pi_level=pi_level)
+
+        self.S_star_se, self.N_star_se = get_active_null_coalition_list(self.sdp_index_se, self.size_se)
+        self.sdp_rules, self.rules, self.sdp_all, self.rules_data, self.w_rules = self.acv_explainer.compute_sdp_maxrules(
+            x_rule, y_rule,
+            self.x_train, self.y_train, self.S_star_se, verbose=True)
+
+    def run_regional_divergent_set(self, stop=True, pi_level=0.8):
+        print('### Computing the regional divergent set of (x, y)')
+
+        self.ddp_importance_regional, self.ddp_index_regional, self.size_regional, self.ddp_regional = \
+            self.acv_explainer.importance_ddp_intv(self.x_rules, self.y_rules,
+                                                   self.x_train, self.y_train,
+                                                   self.rules,
+                                                   stop=stop,
+                                                   pi_level=pi_level)
+
+        self.S_star_regional, self.S_bar_set_regional = \
+            acv_explainers.utils.get_active_null_coalition_list(self.ddp_index_regional, self.size_regional)
+        self.ddp_regional, self.w_regional = self.acv_explainer.compute_ddp_weights(self.x_rules, self.y_rules,
+                                                                                    self.x_train, self.y_train,
+                                                                                    S=self.S_bar_set_regional)
+
+    def run_regional_counterfactual_rules(self, acc_level=0.8, pi_level=0.8):
+        print('### Computing the regional counterfactual rules of (x, y)')
+
+        self.counterfactuals_samples_regional, self.counterfactuals_samples_sdp_regional, \
+        self.counterfactuals_samples_w_regional = \
+            return_ge_global_counterfactuals(self.acv_explainer, self.x_rules, self.y_rules,
+                                             self.S_star_regional, self.S_bar_set_regional,
+                                             self.x_train, self.y_train, self.w_regional, acc_level, self.rules,
+                                             pi_level)
+
+    def run_sampling_regional_counterfactuals(self, max_obs=2, batch=1000, max_iter=1000, temp=0.5):
+        print('### Sampling using the regional counterfactual rules')
+
+        outlier_score = lambda x: self.isolation.decision_function(x)
+
+        self.dist_regional = []
+        self.score_regional = []
+        self.errs_regional = []
+
+        for i in range(self.x_rules.shape[0]):
+            if len(self.counterfactuals_samples_regional[i]) != 0:
+                x_in = np.prod([(self.x_test[:, s] <= self.rules[i, s, 1]) * (self.x_test[:, s] > self.rules[i, s, 0])
+                                for s in range(self.x_train.shape[1])], axis=0).astype(bool)
+                print('observations in rule = {}'.format(np.sum(x_in)))
+                if np.sum(x_in) > 0:
+                    for xi in tqdm(self.x_test[x_in][:max_obs]):
+                        a, sco = simulated_annealing(outlier_score, xi, self.S_star_regional[i], self.x_train,
+                                                     self.counterfactuals_samples_regional[i][
+                                                         np.argmax(self.counterfactuals_samples_sdp_regional[i])][0],
+                                                     batch, max_iter, temp)
+                        self.dist_regional.append(np.squeeze(a))
+                        self.score_regional.append(sco)
+                        self.errs_regional.append(self.acv_explainer.predict(
+                            self.dist_regional[-1].reshape(1, -1)) != self.acv_explainer.predict(xi.reshape(1, -1)))
+
+        self.accuracy_regional = np.mean(self.errs_regional)
+        self.coverage_regional = len(self.errs_regional) / self.x_rules.shape[0]
+
+    def show_global_counterfactuals(self):
+
+        for idt in range(self.rules.shape[0]):
+            print('Example {}'.format(idt))
+            r = self.rules[idt]
+
+            print_rule(self.columns_name, r, self.sdp_rules[idt], True, self.y_rules[idt])
+
+            print('  ')
+            print('  ')
+            print('  ')
+            for l in range(len(self.counterfactuals_samples_sdp_regional[idt])):
+                print('Example {} - Counterfactual {}'.format(idt, l))
+                print_rule(self.columns_name, self.counterfactuals_samples_regional[idt][l][0],
+                           self.counterfactuals_samples_sdp_regional[idt][l], False, False)
+                print('  ')
+            print(' ')
+
+    def show_local_counterfactuals(self, x, y):
+
+        for idt in range(x.shape[0]):
+            print(self.columns_name)
+            print('Example {} = {}'.format(idt, x[idt]))
+
+            print('  ')
+            print('  ')
+            print('  ')
+            for l in range(len(self.counterfactuals_samples_sdp_local[idt])):
+                print('Example {} - Counterfactual {}'.format(idt, l))
+                print_rule(self.columns_name, self.counterfactuals_samples_local[idt][l][0],
+                           self.counterfactuals_samples_sdp_local[idt][l], False, False)
+                print('  ')
+            print(' ')
\ No newline at end of file
